{% extends "base.html" %}

{% block content %}

<div class="static-page">
  <div class="page-content">
    <div class="page-content__header">
      <div class="page-content__title">{{ page.title }}</div>
    </div>

    <div class="page-content__body">
      <div class="page-body">
    {{ page.content }}
      </div>
    </div>
  </div>
</div>
{% endblock %}
