/* dl gets used both for defining each top-level `.. autofunc` on a page (where we want vertical margsin)
   and is wrapped around multiline signatures (where we don't).
   If a dl is being used inside a .sig, that's a multiline signature; remove its margins. */
.sig > dl {
  margin-block-start: 0rem;
  margin-block-end: 0rem;
}

/* with thanks to https://github.com/pradyunsg/furo/discussions/749 */
.sig:not(.sig-inline) {
  padding-left: 0.5em;
  text-indent: 0;
}
