:root {
  --primary-h: 217;
  --primary-s: 100%;
  --primary-l: 19%;

  --text-h: 0;
  --text-s: 0%;

  --color-primary: hsl(var(--primary-h), var(--primary-s), var(--primary-l));
  --link-color: hsl(var(--primary-h), 100%, 35%);
  --font-family: 'Lato', -apple-system, BlinkMacSystemFont, "Segoe UI", Roboto, "Helvetica Neue", Arial, sans-serif;
  --text-color: hsl(var(--text-h), var(--text-s), 2%);

  --nav-link-hover-color: #e9ecef
}

/* start css reset */

*,
*::before,
*::after {
  box-sizing: border-box;
}

* {
  margin: 0;
}

input,
button,
textarea,
select {
  font: inherit;
}
/* end css reset */

body {
  font-family: var(--font-family);
  background-color: white;
}

h1,
h2,
h3 {
  line-height: 1.2;
}

code {
  background-color: #f5f5f5;
  padding: 0.058rem 0.25rem;
  border-radius: 4px;
  color: #c7254e;
  font-size: 0.92em;
}

pre {
  background-color: #f5f5f5;
  padding: 0.9rem;
  border-radius: 6px;
  overflow-x: auto;
  margin: 1.5rem 0;
  border: 1px solid #ddd;
  line-height: 1.1;
}

pre code {
  padding: 0;
  color: var(--text-color);
  font-size: 0.85rem;
}

@media only screen and (max-width: 25em) {
  .profile {
    max-width: 100%;
  }
}

@media print {
  body {
    max-width: none;
  }
}

.page {
  font-family: var(--font-family);
  background-color: white;
  color: var(--text-color);
  line-height: 1.6;
}

.page__container {
  min-height: 100vh;
  display: flex;
  flex-direction: column;
}

.header {
  background-color: #f8f9fa;
  padding: 0.5rem 0;
  color: var(--text-color);
  border-bottom: 1px solid #dee2e6;
}

.header__container {
  max-width: 1200px;
  margin: 0 auto;
  padding: 0 1rem;
  display: flex;
  justify-content: space-between;
  align-items: center;
  flex-wrap: wrap;
  gap: 1rem;
}

.header__brand {
  flex-shrink: 0;
}

.brand__link {
  display: flex;
  align-items: center;
  gap: 0.3rem;
  text-decoration: none;
  color: inherit;
  padding: 0rem 0.4rem;
  border-radius: 4px;
  transition: color 0.05s ease;
}

.brand__link:hover {
  background-color: var(--nav-link-hover-color);
  border-radius: 4px;
}

.brand__logo {
  flex-shrink: 0;
}

.brand__image {
  width: 80px;
  height: 80px;
  display: block;
}

.brand__text {
  display: flex;
  flex-direction: column;
}

.brand__title {
  font-size: 1.5rem;
  font-weight: 600;
  line-height: 1.2;
  color: var(--color-primary);
}

.brand__subtitle {
  font-size: 1rem;
  font-weight: 400;
  color: hsl(var(--primary-h), 0%, 35%);
}

.nav__list {
  display: flex;
  list-style: none;
  margin: 0;
  padding: 0;
}

.nav__link {
  color: var(--link-color);
  text-decoration: none;
  font-weight: 600;
  font-size: 1.15rem;
  padding: 0.75rem 1rem;
  border-radius: 4px;
}

.nav__link:hover {
  color: hsl(var(--primary-h), var(--primary-s), 20%);
  background-color: var(--nav-link-hover-color);
  border-radius: 4px;
}

.main {
  flex: 1;
  padding: 3rem 0;
}

.main__container {
  max-width: 1200px;
  margin: 0 auto;
  padding: 0 1rem;
}

.home-page {
  max-width: 850px;
  margin: 0 auto;
}

.main-page__header {
  text-align: center;
  margin-bottom: 3rem;
}

.main-page__header-text {
  font-size: 4.5rem;
  color: var(--color-primary);
  line-height: 1.1;
  margin: 0;
}

.newsletter-box {
  background-color: #f8f9fa;
  border: 1px solid #dee2e6;
  border-radius: 8px;
  padding: 2rem;
  margin-top: 2rem;
}

.newsletter-box h2 {
  margin-top: 0;
  margin-bottom: 1rem;
  color: var(--color-primary);
}

.newsletter-box p {
  margin-bottom: 1.5rem;
}

.newsletter-form__content {
  display: flex;
  gap: 1rem;
  flex-wrap: wrap;
}

.newsletter-form__input {
  padding-left: 1rem;
  padding-right: 1rem;
  border: 1px solid #ddd;
  border-radius: 6px;
  font-size: 1rem;
  flex: 1;
  min-width: 250px;
  max-width: 350px;
}

.newsletter-form__input:focus {
  outline: none;
  border-color: var(--color-primary);
  box-shadow: 0 0 0 2px hsl(var(--primary-h), var(--primary-s), var(--primary-l), 0.1);
}

.newsletter-form__button {
  padding: 0.75rem 1.5rem;
  background-color: var(--color-primary);
  color: white;
  border: none;
  border-radius: 4px;
  font-size: 1rem;
  font-weight: 500;
  cursor: pointer;
  transition: background-color 0.2s ease;
}

.newsletter-form__button:hover {
  background-color: hsl(var(--primary-h), var(--primary-s), 10%);
}

.articles-page {
  max-width: 1000px;
  margin: 0 auto;
}

.articles-page__header {
  text-align: center;
  margin-bottom: 2rem;
}

.articles-page__title {
  font-size: 2.5rem;
  font-weight: 600;
  color: var(--color-primary);
  margin-bottom: 0.5rem;
}

.articles-page__subtitle {
  font-size: 1.125rem;
  color: #6c757d;
  line-height: 1.5;
}

.articles-grid {
  display: flex;
  flex-direction: column;
  gap: 2rem;
}

.articles__list {
  display: flex;
  flex-direction: column;
  gap: 2rem;
}

.article-card {
  background-color: white;
  border-radius: 8px;
  border: 1px solid #ddd;
  padding-top: 2rem;
  padding-left: 2rem;
  padding-right: 2rem;
  padding-bottom: 1rem;
  transition: transform 0.05s ease, box-shadow 0.05s ease;
  cursor: pointer;
  box-shadow: 0 1.5px 4.5px rgba(0, 0, 0, 0.1);
}

.article-card:hover {
  box-shadow: 0 4px 12px rgba(0, 0, 0, 0.1);
}

.article-card__header {
  margin-bottom: 0.75rem;
}

.article-card__title {
  font-size: 1.5rem;
  font-weight: 600;
  line-height: 1.3;
  color: var(--color-primary);
}


.article-card__content ul,
.article-card__content ol {
  margin-bottom: 1rem;
  padding-left: 2rem;
}

.article-card__content li {
  margin-bottom: 0.25rem;
}

.article-card__content a {
  color: var(--link-color);
  text-decoration: none;
}


.article-card__content p {
  margin-bottom: 1rem;
}

.article-card__content a:hover {
  text-decoration: underline;
}

.article-card-meta {
  display: flex;
  gap: 0.5rem;
  flex-wrap: wrap;
  align-items: center;
  margin-bottom: 0.75rem;
}

.article-card-meta__separator {
  font-size: 0.75rem;
  color: var(--color-primary);
}

.article-card-meta__item {
  display: flex;
  gap: 0.25rem;
  align-items: center;
}

.article-card-meta__value {
  font-size: 1rem;
  color: var(--color-primary);
}

.article-card-meta__value a {
  color: inherit;
  text-decoration: none;
}

.article-card-meta__value a:hover {
  text-decoration: underline;
}

.article-page {
  max-width: 850px;
  margin: 0 auto;
}

.article__header {
  margin-bottom: 2rem;
  padding-bottom: 2rem;
  border-bottom: 2px solid #e2e2e2;
}

.article__title {
  font-size: 2.5rem;
  font-weight: 600;
  line-height: 1.2;
  margin-bottom: 1.5rem;
  color: var(--color-primary);
}

.article-meta {
  display: flex;
  gap: 0.5rem;
  flex-wrap: wrap;
  align-items: center;
}

.article-meta__separator {
  font-size: 0.75rem;
  color: var(--color-primary);
}

.article-meta__item {
  display: flex;
  gap: 0.25rem;
  align-items: center;
}

.article-meta__value {
  font-size: 1.2rem;
  color: var(--color-primary);
}

.article-meta__value a {
  color: inherit;
  text-decoration: none;
}

.article-meta__value a:hover {
  text-decoration: underline;
}

.article-content {
  line-height: 1.6;
  font-size: 1.125rem;
  color: var(--text-color);
}

.article-content h1,
.article-content h2,
.article-content h3,
.article-content h4,
.article-content h5,
.article-content h6 {
  margin-top: 2rem;
  margin-bottom: 1rem;
  line-height: 1.3;
  color: var(--color-primary);
}

.article-content h1 {
  font-size: 2rem;
}

.article-content h2 {
  font-size: 1.5rem;
}

.article-content h3 {
  font-size: 1.25rem;
}

.article-content p {
  margin-bottom: 1rem;
}

.article-content ul,
.article-content ol {
  margin-bottom: 1rem;
  padding-left: 2rem;
}

.article-content li {
  margin-bottom: 0.25rem;
}

.article-content a {
  color: var(--link-color);
  text-decoration: none;
}

.article-content a:hover {
  text-decoration: underline;
}

.article-content hr {
  margin-top: 1.5rem;
  margin-bottom: 1.5rem;
  border: 1px solid #e2e2e2;
}

.cta-buttons {
  display: flex;
  gap: 1.5rem;
  justify-content: center;
  align-items: center;
  margin: 2rem 0;
}

.cta-button {
  display: inline-flex;
  align-items: center;
  gap: 0.75rem;
  padding: 0.8rem 1.2rem;
  background-color: hsl(var(--primary-h), 80%, 56%);
  color: white !important;
  text-decoration: none !important;
  font-weight: 600;
  font-size: 1.1rem;
  border: none;
  border-radius: 6px;
  box-shadow: 0 2px 8px rgba(0, 0, 0, 0.15);
  text-align: center;
  cursor: pointer;
}

.cta-icon {
  width: 1.2rem;
  height: 1.2rem;
  filter: brightness(0) invert(1);
}

.cta-button:hover {
  background-color: hsl(var(--primary-h), 80%, 51%);
  box-shadow: 0 4px 12px rgba(0, 0, 0, 0.2);
}

.footnote {
  padding-top: 1.25rem;
}

.static-page {
  max-width: 850px;
  margin: 0 auto;
}

.page-content__header {
  margin-bottom: 2rem;
  padding-bottom: 2rem;
  border-bottom: 1px solid #eee;
}

.page-content__title {
  font-size: 2.5rem;
  font-weight: 600;
  line-height: 1.2;
  color: var(--color-primary);
}

.page-body {
  line-height: 1.7;
  color: var(--text-color);
}

.page-body h1,
.page-body h2,
.page-body h3,
.page-body h4,
.page-body h5,
.page-body h6 {
  margin-top: 2rem;
  margin-bottom: 1rem;
  line-height: 1.3;
  color: var(--color-primary);
}

.page-body h1 {
  font-size: 2rem;
}

.page-body h2 {
  font-size: 1.5rem;
}

.page-body h3 {
  font-size: 1.25rem;
}

.page-body p {
  margin-bottom: 1rem;
}

.page-body ul,
.page-body ol {
  margin-bottom: 1rem;
  padding-left: 2rem;
}

.page-body li {
  margin-bottom: 0.25rem;
}

.page-body a {
  color: var(--link-color);
  text-decoration: none;
}

.page-body a:hover {
  text-decoration: underline;
}

.footer {
  background-color: #f8f8f8;
  border-top: 1px solid #ddd;
  margin-top: auto;
}

.footer__container {
  max-width: 1200px;
  margin: 0 auto;
  padding: 1rem;
}

.footer__content {
  text-align: center;
}

.footer__text {
  color: #666;
  font-size: 0.875rem;
}

.separator {
  border-bottom: 2px solid #cccccc;
}

@media (max-width: 768px) {
  .header__container {
    flex-direction: column;
    text-align: center;
  }

  .nav__list {
    flex-wrap: wrap;
    justify-content: center;
  }

  .brand__link {
    flex-direction: column;
    text-align: center;
  }

  .article__title {
    font-size: 2rem;
  }

  .article-card-meta,
  .article-meta {
    flex-direction: column;
    gap: 0.5rem;
  }

  .newsletter-form__content {
    flex-direction: column;
    align-items: center;
  }

  .newsletter-form__input {
    max-width: 100%;
  }

  .articles__title {
    font-size: 1.75rem;
  }

  .cta-buttons {
    flex-direction: column;
    gap: 1rem;
  }
}
